import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.suitable.sara.api.robot_router import RobotRouter
from ablelabs.neon.common.suitable.struct import Location, Speed, FlowRate


class MotionAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.motion_initialize)
    async def initialize(self):
        pass

    @run_server_func(RobotRouter.motion_delay)
    async def delay(self, sec: float):
        pass

    # @run_server_func(Router.motion_move_to_ready)
    # async def move_to_ready(self):
    #     pass

    @run_server_func(RobotRouter.motion_move_to)
    async def move_to(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_pick_up_tip)
    async def pick_up_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_drop_tip)
    async def drop_tip(
        self,
        pipette_number: list[int],
        location: Location,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_rise_tip)
    async def rise_tip(
        self,
        pipette_number: list[int],
        height_offset: float | list[float],
        z_speed: Speed | list[Speed],
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_aspirate)
    async def aspirate(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        blow_out_flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_dispense)
    async def dispense(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        location: Location = None,
        flow_rate: FlowRate = None,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_mix)
    async def mix(
        self,
        pipette_number: list[int],
        volume: float | list[float],
        iteration: int,
        location: Location = None,
        flow_rate: FlowRate = None,
        delay: float = 0.0,
        optimize: bool = False,
    ):
        pass

    @run_server_func(RobotRouter.motion_blow_out)
    async def blow_out(
        self,
        pipette_number: list[int],
        flow_rate: FlowRate = None,
    ):
        pass



def merge_dict(original_dict: dict, new_dict: dict):
    # 동일한 key가 있는 경우, original_dict의 내부 dict을 유지하면서 새로운 dict을 병합.
    for key, value in new_dict.items():
        if (
            key in original_dict
            and isinstance(value, dict)
            and isinstance(original_dict[key], dict)
        ):
            merge_dict(original_dict[key], value)
        else:
            original_dict[key] = value


def get_nested_value(d: dict, keys: list):
    """
    중첩된 딕셔너리에서 키를 순서대로 조회하여 값을 반환하는 함수.
    키가 존재하지 않으면 None을 반환.

    Args:
        d (dict): 딕셔너리
        keys (list): 키의 리스트

    Returns:
        중첩된 값 또는 None
    """
    for key in keys:
        if isinstance(d, dict) and key in d:
            d = d[key]
        else:
            return None
    return d



import datetime
import socket
import threading
import json
from loguru import logger
import asyncio

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.base_tcp import *


class TcpClient(BaseTcp):
    def __init__(
        self,
        name="Client",
        encoding: str = "utf-8",
        log_func: Callable = logger.trace,
    ):
        super().__init__(name=name, encoding=encoding, log_func=log_func)

    async def connect(self, ip, port):
        self._reader, self._writer = await asyncio.open_connection(ip, port)
        self._is_connected = True
        try:
            asyncio.create_task(
                self.handle_receive(), name=self.handle_receive.__name__
            )
        except Exception as e:
            logger.exception(f"{e}")
        for on_connected in self.on_connected:
            await on_connected()


# if __name__ == "__main__":
#     TcpClient.example()



def check_com_port_count(required: int, port: str):
    import serial.tools.list_ports

    device_ports = [comport.device for comport in serial.tools.list_ports.comports()]
    if (count := len(device_ports)) < required:
        raise Exception(f"COM port count : {count} (< {required})")
    config_port = port
    if config_port == "any":
        # port 번호 오름차순으로 정렬.
        device_ports.sort(key=lambda port: int(port[3:]), reverse=False)
        return device_ports
    else:
        config_ports = config_port.split(" ")
        if set(config_ports) > set(device_ports):
            raise Exception(
                f"COM port mismatch : device manager={device_ports}  config={config_port}"
            )
        return config_ports



from enum import Enum, IntFlag, auto
from enum import IntEnum as BaseIntEnum
from typing import Callable, Generic, TypeVar, Any


class StrEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    @classmethod
    def from_str(cls, value: str):
        return cls[value.upper()]


class IntEnum(BaseIntEnum):
    def __str__(self):
        return f"{self.name}({self.value})"


class ValueEnum(str, Enum):
    def _generate_next_value_(name, start, count, last_values):
        return name

    def __repr__(self):
        return self.value

    def __str__(self):
        return str(self.value)


T = TypeVar("T")


class ItemGetable(property, Generic[T]):
    def __init__(self, wrapped: Callable[[T], Any]):
        super().__init__(lambda s: self)
        self.wrapped: Callable[[T], Any] = wrapped

    def __getitem__(self, item: T) -> Any:
        return self.wrapped(item)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.suitable.enums import Axis

PIPETTE_COUNT: int = 8
PIPETTE_NUMBERS = [i + 1 for i in range(PIPETTE_COUNT)]

AXIS_YS = [Axis.Y[pipette_number] for pipette_number in PIPETTE_NUMBERS]
AXIS_ZS = [Axis.Z[pipette_number] for pipette_number in PIPETTE_NUMBERS]
AXIS_PS = [Axis.P[pipette_number] for pipette_number in PIPETTE_NUMBERS]

Y_MIN_INTERVAL_MM = 9.0

SLEEP_CHECK_COM_PORT = 3
SLEEP_CHECK_HW_STATUS = 1
SLEEP_UPDATE_DIO_LOOP = 0.5
SLEEP_LOOP = 0.1
SLEEP_AFTER_TIMEOUT = 5


from dataclasses import dataclass
import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.structs import Speed, FlowRate, LedBarParam
from ablelabs.neon.common.suitable.enums import LocationType, LocationReference


@dataclass
class Location:
    location_type: LocationType
    location_number: int
    well: list[str]
    reference: LocationReference
    offset: tuple[float] = (0, 0, 0)


def location(
    location_type: LocationType,
    location_number: int = None,
    well: list[str] = None,
    reference: LocationReference = None,
    offset: tuple[float] = (0, 0, 0),
):
    return Location(
        location_type=location_type,
        location_number=location_number,
        well=well,
        reference=reference,
        offset=offset,
    )



import time
from typing import Callable
import asyncio
import inspect
from functools import update_wrapper


def log_elapsed_time(log_func: Callable):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            elapsed_time = time.perf_counter() - start_time
            log_func(f"{func.__name__} : {elapsed_time*1e3:.1f} ms")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


def log_elapsed_time_async(log_func: Callable):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            start_time = time.perf_counter()
            result = await func(*args, **kwargs)
            elapsed_time = time.perf_counter() - start_time
            log_func(f"{func.__name__} : {elapsed_time*1e3:.1f} ms")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


@log_elapsed_time(print)
def test_check_time(n):
    for _ in range(n):
        time.sleep(0.01)


@log_elapsed_time_async(print)
async def test_check_time_async(n):
    for _ in range(n):
        await asyncio.sleep(0.01)


def get_args_dict(func: Callable, *args, **kwargs):
    sig = inspect.signature(func)
    try:  # sypark 여기 try문을 다른 걸로 대체할 수 없을까?
        bound = sig.bind(*args, **kwargs)
    except:
        bound = sig.bind(None, *args, **kwargs)
    # if "self" in sig.parameters:
    #     bound = sig.bind(None, *args, **kwargs)
    # else:
    #     bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    result = dict(bound.arguments)
    if "self" in result and result["self"] is None:
        del result["self"]
    return result


def get_dict_str(value: dict):
    args_strs = []
    for key, value in value.items():
        if isinstance(value, list):
            value_str = [str(v) for v in value]
            args_strs.append(f"{key}=[{', '.join(value_str)}]")
        else:
            args_strs.append(f"{key}={value}")
    result = f"{' '.join(args_strs)}"
    return result


def get_func_args_str(func: Callable, args_dict: dict):
    # args_strs = [f"{k}={v}" for k, v in args_dict.items() if k != "self"]
    args_strs = []
    for key, value in args_dict.items():
        if key == "self":
            continue
        if isinstance(value, list):
            value_str = [str(v) for v in value]
            args_strs.append(f"{key}=[{', '.join(value_str)}]")
        elif inspect.iscoroutine(value):
            args_strs.append(f"{key}={value.__qualname__}")
        elif isinstance(value, asyncio.Task):
            args_strs.append(f"{key}={value.get_coro().__qualname__}")
        else:
            args_strs.append(f"{key}={value}")
    result = f"{func.__name__}({' '.join(args_strs)})"
    return result


def log_func_args(
    log_before_func: Callable = None,
    log_after_func: Callable = None,
):
    def decorator(func: Callable):
        def wrapper(*args, **kwargs):
            args_dict = get_args_dict(func, *args, **kwargs)
            func_args_strs = get_func_args_str(func, args_dict)
            if log_before_func:
                log_before_func(f"{func_args_strs}")
            result = func(*args, **kwargs)
            if log_after_func:
                log_after_func(f"{func_args_strs} -> {result}")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


def log_func_args_async(
    log_before_func: Callable = None,
    log_after_func: Callable = None,
):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            args_dict = get_args_dict(func, *args, **kwargs)
            func_args_strs = get_func_args_str(func, args_dict)
            if log_before_func:
                log_before_func(f"{func_args_strs}")
            result = await func(*args, **kwargs)
            if log_after_func:
                log_after_func(f"{func_args_strs} -> {result}")
            return result

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


# @log_elapsed_time_async(print)
@log_func_args_async(log_before_func=print, log_after_func=print)
async def test_log_func_args_async(a: int, b: float, c: str):
    await asyncio.sleep(0.01)


def skip_func_async(skip_func: Callable[[], bool], return_value=None):
    def decorator(func):
        async def wrapper(*args, **kwargs):
            if skip_func():
                return return_value
            return await func(*args, **kwargs)

        update_wrapper(wrapper, func)
        return wrapper

    return decorator


@skip_func_async(skip_func=lambda: True, return_value="return_sample")
async def test_skip_func_async_do_skip():
    print(f"{test_skip_func_async_do_skip.__name__}")


@skip_func_async(skip_func=lambda: False, return_value="return_sample")
async def test_skip_func_async_donot_skip():
    print(f"{test_skip_func_async_donot_skip.__name__}")


async def main():
    # test_check_time(1)
    # test_check_time(2)
    # test_check_time(3)

    # await test_check_time_async(1)
    # await test_check_time_async(2)
    # await test_check_time_async(3)

    # await test_log_func_args_async(a=1, b=2.3, c="abc")
    # await test_log_func_args_async(a=0, b=0, c="")
    # await test_log_func_args_async(1, 2.3, "abc")

    print(await test_skip_func_async_do_skip())
    print(await test_skip_func_async_donot_skip())


if __name__ == "__main__":
    asyncio.run(main())
